from PIL import Image, ImageDraw

from .basegraph import BaseGraph
from .utils import circle_xy, limit


class PieChart(BaseGraph):
    """Class representing a Pie chart."""
    def __init__(self,
                 *,
                 radius: int,
                 width: int = 0,
                 angle: int = 0,
                 emboss: int = 0,
                 space_between: int = 0) -> None:
        """
        PieChart constructor.

        Attributes
        ----------
        radius: `int`
            Radius of the chart circle.
        width: `int`
            If None, graph will be pie-shaped.
            Otherwise, graph will be donut-shaped with specified thickness.
        angle: `int`
            Start angle of the chart.
        emboss: `int`
            If = 0, graph will be flat. 
            Otherwise, graph parts will be different size based on value.
            If < 0, parts with higher weight will be smaller.
        space_between: `int`
            Space between graph parts.
        """
        super().__init__()

        self.radius = radius
        self.width = width
        self.angle = angle
        self.emboss = emboss
        self.space_between = space_between

    @property
    def radius(self) -> int:
        """Chart circle radius."""
        return self._radius
    
    @radius.setter
    def radius(self, value: int):
        if (hasattr(self, '_thickness')
            and self.width > value):
            raise ValueError("'radius' can not be smaller than 'thickness'")
        self._radius = value

    @property
    def width(self) -> int:
        """Width of the donut-shaped graph."""
        return self._width
    
    @width.setter
    def width(self, value: int):
        if (hasattr(self, '_radius') 
            and self.radius < value):
            raise ValueError("'thickness' can not be bigger than 'radius'")
        if (hasattr(self, '_emboss') 
            and abs(self.emboss) * 2 < value):
            raise ValueError("'thickness' can not be bigger than absolute value of 'emboss' twice")
        self._width = value

    @property
    def angle(self) -> int:
        """Chart start angle"""
        return self._angle
    
    @angle.setter
    def angle(self, value: int):
        self._angle = value

    @property
    def emboss(self) -> int:
        """Graph parts max emboss."""
        return self._emboss
    
    @emboss.setter
    def emboss(self, value: int):
        if (hasattr(self, '_thickness') 
            and self.width < abs(value) * 2):
            raise ValueError("'emboss' can not be bigger than half of 'thickness'")
        self._emboss = value

    @property
    def space_between(self) -> int:
        """Space between chart parts."""
        return self._space_between
    
    @space_between.setter
    def space_between(self, value: int):
        self._space_between = value

    def draw(self) -> Image.Image:
        image = Image.new('RGBA', (self.radius * 2,)*2)
        num_items = len(self.items)

        if num_items == 0:
            return image
        
        weights = [item.weight for item in self.items]
        total_weight = sum(weights)
        start_angle = self.angle
        width = self.width
        emboss = self.emboss
        template = image.copy()

        offsets = limit(
            weights, 
            0 if emboss < 0 else abs(emboss), 
            abs(emboss) if emboss < 0 else 0)

        for i, item in enumerate(self.items):
            img = template.copy()
            draw = ImageDraw.Draw(img)
            offset = offsets[i]

            if total_weight == 0: 
                angle = 360 / num_items
            else:
                angle = 360 * item.weight / total_weight

            if item.color is not None:
                draw.pieslice(((0 + offset,)*2, (img.width - offset,)*2),
                            start_angle, start_angle + angle,
                            fill=item.color.rgba)
                
                if self.space_between > 0:
                    draw.line(((self.radius,)*2, circle_xy(self.radius, self.radius, start_angle + angle)),
                            fill=(0, 0, 0, 0), width=self.space_between)
                    draw.line(((self.radius,)*2, circle_xy(self.radius, self.radius, start_angle)),
                            fill=(0, 0, 0, 0), width=self.space_between)
                    draw.ellipse(((self.radius - self.space_between / 2,)*2, (self.radius + self.space_between / 2,)*2),
                                fill=(0, 0, 0, 0), width=0)
            
                if width > 0:
                    draw.ellipse(((width - offset,)*2, (self.radius * 2 - width + offset,)*2),
                                fill=(0, 0, 0, 0))
            
                image.alpha_composite(img)

            start_angle += angle

        return image


    


class Color:
    """Class representing a 8-bit color."""
    def __init__(self,
                 color: int | str | tuple[int, int, int] | tuple[int, int, int, int] | None = None,
                 /,
                 alpha: int | None = None) -> None:
        """
        Constructs `Color` from value.

        Attributes
        ----------
        color: `int` | `str` | `tuple[int, int, int]` | `tuple[int, int, int, int]`
            Color value. Can be `HEX`, `RGB`, `RGBA` or color number.
            If `None`, sets to black transparent color.
        """        
        self.alpha = 255

        _err = ValueError("invalid color value")

        if isinstance(color, str):
            from .utils import hex_to_rgb
            color = hex_to_rgb(color.removeprefix("#"))
        elif isinstance(color, int):
            from .utils import num_to_rgb
            color = num_to_rgb(color)
        elif isinstance(color, (tuple, list)):
            if len(color) == 4:
                if alpha is None:
                    self.alpha = color[3]
                color = color[:3]
            elif len(color) != 3:
                raise _err
        elif color is None:
            self.alpha = 0
            color = (0, 0, 0)
        else:
            raise _err
    
        self._rgb: tuple[int, int, int] = color

    def __bool__(self):
        return all(self.rgba)

    @property
    def rgb(self) -> tuple[int, int, int]:
        """Get the color in `RGB` format."""
        return self._rgb
    
    @property
    def rgba(self) -> tuple[int, int, int, int]:
        """Get the color in `RGBA` format."""
        return self._rgb + (self._alpha,)

    @property
    def hex(self) -> str:
        """Get the color in `HEX` format."""
        from .utils import rgb_to_hex
        return rgb_to_hex(self._rgb)
    
    @property
    def number(self) -> int:
        """Get the color as number."""
        from .utils import rgb_to_num
        return rgb_to_num(self._rgb)
    
    @property
    def alpha(self) -> int:
        """Get the color alpha."""
        return self._alpha
    
    @alpha.setter
    def alpha(self, value: int):
        self._alpha = value


class Size:
    def __init__(self, size: tuple[int, int]) -> None:
        if (len(size) != 2
            or not all((isinstance(i, int) for i in size))):
            raise ValueError("'size' must be iterable containing 2 integers")
       
        self._size: tuple[int, int] = size
    
    def __getitem__(self, key: int):
        return self._size[key]
    
    def __setitem__(self, key: int, value: int):
        self._validate(value, 'value')
        self._size[key] = value

    @staticmethod
    def _validate(value: int, name: str) -> None:
        if (not isinstance(value, int)
            or value < 0):
            raise ValueError(f"'{name}' must be greater than 0")
    
    @property
    def width(self):
        """Width."""
        return self._size[0]
    
    @width.setter
    def width(self, value: int):
        self._validate(value, 'width')
        self._size[0] = value

    @property
    def height(self):
        """Height."""
        return self._size[1]
    
    @height.setter
    def height(self, value: int):
        self._validate(value, 'height')
        self._size[1] = value

    


from PIL import Image, ImageDraw

from .basegraph import BaseGraph
from .color import Color
from .size import Size
from .utils import get_color, limit, interpolate, Interpolation


class LineChart(BaseGraph):
    """Class representing a Line chart."""
    def __init__(self,
                 *,
                 size: Size | tuple[int, int],
                 thickness: int = 1,
                 fill: Color | int | str | tuple[int, int, int] | tuple[int, int, int, int] | None = ...,
                 outline: Color | int | str | tuple[int, int, int] | tuple[int, int, int, int] | None = ...,
                 point_width: int = 0,
                 all_points: bool = False,
                 num_points: int = 0,
                 interpol: Interpolation = Interpolation.LINEAR,
                 min_height: int = 0) -> None:
        """
        LineChart constructor.

        Attributes
        ----------
        size: `Size` | `tuple[int, int]`
            Graph image size.
        thickness: `int`
            Line thickness.
        fill: `Color`
            Fill color.
        outline: `Color`
            Line color.
        point_width: `int`
            Point width.
        all_points: `bool`
            If `True`, all points (including intermediate ones) will be drawn.
            Otherwise, only source points will be displayed.
        num_points: `int`
            Number of points. If <= 0, equals to the number of items.
        interpol: `Interpolation`
            Kind of interpolation. Used to make a smooth curve.
        min_height: `int`
            Minimum height from bottom of the graph.
        """
        super().__init__()

        self.size = size
        self.thickness = thickness
        self.fill = get_color(fill)
        self.outline = get_color(outline)
        self.point_width = point_width
        self.all_points = all_points
        self.num_points = num_points
        self.interpol = interpol
        self.min_height = min_height

    @property
    def size(self) -> Size:
        """Graph size."""
        return self._size
    
    @size.setter
    def size(self, value: Size | tuple[int, int]):
        if not isinstance(value, Size): 
            value = Size(value)

        self._size = value

    @property
    def thickness(self) -> int:
        """Stroke thickness."""
        return self._thickness
    
    @thickness.setter
    def thickness(self, value: int):
        self._thickness = value

    @property
    def fill(self) -> Color:
        """Fill color."""
        return self._fill
    
    @fill.setter
    def fill(self, value: Color):
        self._fill = value

    @property
    def outline(self) -> Color:
        """Outline color."""
        return self._outline
    
    @outline.setter
    def outline(self, value: Color):
        self._outline = value

    @property
    def point_width(self) -> int:
        """Stroke points width."""
        return self._point_width
    
    @point_width.setter
    def point_width(self, value: int):
        self._point_width = value

    @property
    def all_points(self) -> bool:
        """Draw all points or not."""
        return self._all_points
    
    @all_points.setter
    def all_points(self, value: bool):
        self._all_points = value

    @property
    def num_points(self) -> int:
        """Number of points."""
        return self._num_points
    
    @num_points.setter
    def num_points(self, value: int):
        self._num_points = value

    @property
    def interpol(self) -> Interpolation:
        """Kinf of interpolation."""
        return self._interpol
    
    @interpol.setter
    def interpol(self, value: Interpolation):
        self._interpol = value

    @property
    def min_height(self) -> int:
        """Min height (max y value in the image coordinates)."""
        return self._min_height
    
    @min_height.setter
    def min_height(self, value: int):
        if (hasattr(self, '_radius')
            and self.size.height < value):
            raise ValueError("'min_height' can not be bigger than 'size.height'")
        self._min_height = value

    def draw(self) -> Image.Image:
        image = Image.new('RGBA', tuple(self.size))
        num_items = len(self.items)

        if num_items == 0:
            return image
        
        draw = ImageDraw.Draw(image)
        thickness = self.thickness or 1
        num = self.num_points if self.num_points > 0 else num_items
        max_weight = max((i.weight for i in self.items))
        p_radius = self.point_width / 2 if self.point_width > 0 else thickness / 2
        limited_y = limit(
            [max_weight - item.weight for item in self.items], 
            p_radius, 
            self.size.height - p_radius - self.min_height)
        space_between_points = self.size.width / (num_items - 1)
        limited_x = limit(
            [space_between_points * i for i in range(num_items)], 
            p_radius, 
            self.size.width - p_radius)
        
        points = list(zip(limited_x, limited_y))
        smooth_points = interpolate(points, num, kind=self.interpol)

        if self.fill:
            draw.polygon(
                [(p_radius, self.size.height)] 
                + smooth_points 
                + [(self.size.width-p_radius, self.size.height)],
                fill=self.fill.rgba, width=0)

        if self.outline:
            draw.line(smooth_points, fill=self.outline.rgba, width=thickness, joint='curve')

            big_points = (points[0], points[num_items-1])
            if self.point_width:
                big_points = smooth_points if self.all_points else points

            for point in big_points:
                draw.ellipse(
                    (point[0] - p_radius, point[1] - p_radius,
                     point[0] + p_radius, point[1] + p_radius),
                     fill=self.outline.rgba, width=0)

        return image



import math
import enum
import random
import numpy as np
from scipy.interpolate import interp1d


def rgb_to_hex(_rgb: tuple[int, int, int], /):
    return '{:02x}{:02x}{:02x}'.format(_rgb[0], _rgb[1], _rgb[2])


def hex_to_rgb(_hex: str, /):
    return tuple(int(_hex.strip('#')[i:i+2], 16) for i in (0, 2, 4))


def num_to_rgb(_num: int, /):
    return ((_num >> 16) & 255,
            (_num >> 8) & 255,
            _num & 255)


def rgb_to_num(_rgb: tuple[int, int, int], /):
    return (_rgb[0] << 16) + (_rgb[1] << 8) + _rgb[2]


def circle_xy(radius: int, distance: int, angle: int):
    rad = math.radians(angle)
    return (radius + distance * math.cos(rad),
            radius + distance * math.sin(rad))


def get_color(color, /):
    """
    Returns
    -------
    `Color` or `None` if color is `None`.
    """
    from .color import Color

    if isinstance(color, Color):
        return color
    elif color == ...:
        return Color((
            random.randint(0, 255),
            random.randint(0, 255),
            random.randint(0, 255)
        ))
    elif color is not None:
        return Color(color)
    else:
        return Color((0, 0, 0, 0))
    

class Interpolation(enum.Enum):
    """
    Interpolation type.
    """
    LINEAR = 'linear'
    NEAREST = 'nearest'
    NEAREST_UP = 'nearest-up'
    ZERO = 'zero'
    SLINEAR = 'slinear'
    QUADRATIC = 'quadratic'
    CUBIC = 'cubic'
    PREVIUOS = 'previous'
    NEXT = 'next'


def interpolate(
        points: list[tuple[int, int]], 
        num: int | None = None, 
        kind: Interpolation = Interpolation.LINEAR) -> list[tuple[int, int]]:
    """
    Interpolate list of points to make a smooth curve.

    Attributes
    ----------
    points: `list[tuple[int, int]]`
        List of points. Every point must be a tuple containing 2 integers: x and y.
    num: `int` | `None`
        Number of points. If `None`, double the length of the list of points is set.
    kind: `Interpolation`
        The kind of interpolation.
    """
    x, y = zip(*points)
    inter = interp1d(x, y, kind=kind.value)

    if not num:
        num = len(points) * 2

    x_new = np.linspace(min(x), max(x), num)
    y_new = np.clip(inter(x_new), min(y), max(y))

    return list(zip(x_new, y_new))


def limit(
        values: list[int | float],
        minv: int | float,
        maxv: int | float,
        *,
        copy: bool = True) -> np.ndarray:
    """
    Limit array to specific range.

    Attributes
    ----------
    values: `list[int | float]`
        List of values.
    minv: `int | float`
        Minimum (bottom) value.
    maxv: `int | float`
        Maximum (top) value.
    copy: `bool`
        Copy an array or not.
    """
    array = np.array(values, copy=copy)
    _min, _max = min(array), max(array)

    if _max == _min:
        return array

    m = (maxv - minv) / (_max - _min)
    b = maxv - m * _max

    return m * array + b


def linear_to_circle(
        points: list[tuple[int, int]], 
        radius: int, 
        min_radius: int = 0, 
        angle: int = 0) -> list[tuple[int, int]]:
    """
    Convert linear points to circular.

    Attributes
    ----------
    points: `list[tuple[int, int]]`
        List of points.
    radius: `int`
        Max radius.
    min_radius: `int`
        Min radius.
    angle: `int`
        Rotation angle.
    """
    y = [p[1] for p in points]
    max_y = max(y)
    ang = 360 / (len(points) - 1)
    radii = limit([max_y - i for i in y], min_radius, radius)

    return [
        circle_xy(radius, rad, i * ang + angle)
        for i, rad in enumerate(radii)
    ]


from .color import Color
from .utils import get_color


class GraphItem:
    """Class representing a graph item."""
    def __init__(self,
                 *,
                 color: Color | int | str | tuple[int, int, int] | tuple[int, int, int, int] | None = ...,
                 weight: int | float = 1) -> None:
        self.color = get_color(color)
        self.weight = weight

    @property
    def color(self) -> Color | None:
        return self._color
    
    @color.setter
    def color(self, value: Color):
        self._color = get_color(value)

    @property
    def weight(self) -> int | float:
        return self._weight
    
    @weight.setter
    def weight(self, value: int | float):
        self._weight = value
    


from PIL import Image, ImageDraw

from .basegraph import BaseGraph
from .color import Color
from .utils import get_color, interpolate, Interpolation, linear_to_circle


class RadarChart(BaseGraph):
    """Class representing a Radar chart."""
    def __init__(self,
                 *,
                 radius: int,
                 thickness: int = 1,
                 fill: Color | int | str | tuple[int, int, int] | tuple[int, int, int, int] | None = ...,
                 outline: Color | int | str | tuple[int, int, int] | tuple[int, int, int, int] | None = ...,
                 point_width: int = 0,
                 all_points: bool = False,
                 num_points: int = 0,
                 interpol: Interpolation = Interpolation.LINEAR,
                 angle: int = 0,
                 min_radius: int = 0) -> None:
        """
        RadarChart constructor.

        Attributes
        ----------
        radius: `int`
            Radius of the chart shape.
        thickness: `int`
            Line thickness.
        fill: `Color`
            Fill color.
        outline: `Color`
            Line color.
        point_width: `int`
            Point width.
        all_points: `bool`
            If `True`, all points (including intermediate ones) will be drawn.
            Otherwise, only source points will be displayed.
        num_points: `int`
            Number of points. If <= 0, equals to the number of items.
        interpol: `Interpolation`
            Kind of interpolation. Used to make a smooth curve.
        angle: `int`
            Start angle of the chart.
        min_radius: `int`
            Minimum distance between the center and the point.
        """
        super().__init__()

        self.radius = radius
        self.thickness = thickness
        self.fill = get_color(fill)
        self.outline = get_color(outline)
        self.point_width = point_width
        self.all_points = all_points
        self.num_points = num_points
        self.interpol = interpol
        self.angle = angle
        self.min_radius = min_radius

    @property
    def radius(self) -> int:
        """Radar shape radius."""
        return self._radius
    
    @radius.setter
    def radius(self, value: int):
        if (hasattr(self, '_min_radius') 
            and self.min_radius > value):
            raise ValueError("'radius' can not be smaller than 'minimum'")
        self._radius = value

    @property
    def thickness(self) -> int:
        """Stroke thickness."""
        return self._thickness
    
    @thickness.setter
    def thickness(self, value: int):
        self._thickness = value

    @property
    def fill(self) -> Color:
        """Fill color."""
        return self._fill
    
    @fill.setter
    def fill(self, value: Color):
        self._fill = value

    @property
    def outline(self) -> Color:
        """Outline color."""
        return self._outline
    
    @outline.setter
    def outline(self, value: Color):
        self._outline = value

    @property
    def point_width(self) -> int:
        """Stroke points width."""
        return self._point_width
    
    @point_width.setter
    def point_width(self, value: int):
        self._point_width = value

    @property
    def all_points(self) -> bool:
        """Draw all points or not."""
        return self._all_points
    
    @all_points.setter
    def all_points(self, value: bool):
        self._all_points = value

    @property
    def num_points(self) -> int:
        """Number of points."""
        return self._num_points
    
    @num_points.setter
    def num_points(self, value: int):
        self._num_points = value

    @property
    def interpol(self) -> Interpolation:
        """Kinf of interpolation."""
        return self._interpol
    
    @interpol.setter
    def interpol(self, value: Interpolation):
        self._interpol = value

    @property
    def angle(self) -> int | None:
        """Chart start angle."""
        return self._angle
    
    @angle.setter
    def angle(self, value: int):
        self._angle = value

    @property
    def min_radius(self) -> int:
        """Minimum distance between center and point."""
        return self._min_radius
    
    @min_radius.setter
    def min_radius(self, value: int):
        if (hasattr(self, '_radius')
            and self.radius < value):
            raise ValueError("'min_radius' can not be bigger than 'radius'")
        self._min_radius = value

    def draw(self) -> Image.Image:
        image = Image.new('RGBA', (self.radius * 2,)*2)
        items = self.items
        items.append(items[0])
        num_items = len(items)

        if num_items == 0:
            return image
        
        draw = ImageDraw.Draw(image)
        thickness = self.thickness or 1
        num = self.num_points if self.num_points > 0 else num_items
        max_weight = max((i.weight for i in items))
        p_radius = self.point_width / 2 if self.point_width > 0 else thickness / 2
        space_between_points = image.size[0] / (num_items - 1)
      
        points = list(zip(
            [space_between_points * i for i in range(num_items)], 
            [max_weight - item.weight for item in items]))
        smooth_points = interpolate(points, num, kind=self.interpol)
        circle_points = linear_to_circle(
            smooth_points, 
            self.radius - self.point_width, 
            self.min_radius,
            self.angle)

        if self.fill:
            draw.polygon(
                circle_points,
                fill=self.fill.rgba, 
                outline=self.outline.rgba,
                width=0)

        if self.outline:
            draw.line(
                circle_points + [circle_points[0]], 
                fill=self.outline.rgba, 
                width=thickness, 
                joint='curve')

            big_points = (circle_points[0],)
            if self.point_width > 0:
                big_points = circle_points if self.all_points else circle_points[::num//num_items]

            for point in big_points:
                draw.ellipse(
                    (point[0] - p_radius, point[1] - p_radius,
                    point[0] + p_radius, point[1] + p_radius),
                    fill=self.outline.rgba, width=0)

        return image




from PIL import Image

from .item import GraphItem


class BaseGraph:
    def __init__(self) -> None:
        self._items: list[GraphItem] = []
        
    @property
    def items(self) -> list[GraphItem]:
        """Graph items."""
        return self._items.copy()
        
    def add_items(self, *items: GraphItem) -> None:
        """
        Add items to graph.

        Attributes
        ----------
        items: `GraphItem`
            Items to add.

        Raises
        ------
        `ValueError` if item is not of correct type.
        """
        for item in items:
            if not isinstance(item, GraphItem):
                raise ValueError(f"items must be instances of '{GraphItem.__name__}', not {type(item).__name__}")
            self._items.append(item)

    def remove_items(self, *items: GraphItem) -> None:
        """
        Remove items from graph.

        Attributes
        ----------
        items: `GraphItem`
            Items to remove.

        Raises
        ------
        `ValueError` if item is not present.
        """
        for item in items:
            self._items.remove(item)

    def draw(self) -> Image.Image:
        """
        Draw a graph.
        """
        raise NotImplementedError()
